@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2004 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileAccess.cc,v 1.115 2004/04/16 14:27:34 lav Exp $ */
+/* $Id: FileAccess.cc,v 1.133 2006/07/04 05:15:35 lav Exp $ */
 
 #include <config.h>
 
@@ -40,6 +40,7 @@
 #include "netrc.h"
 #include "ArgV.h"
 #include "ConnectionSlot.h"
+#include "SignalHook.h"
 #ifdef WITH_MODULES
 # include "module.h"
 #endif
@@ -48,8 +49,9 @@ FileAccess *FileAccess::chain=0;
 
 void FileAccess::Init()
 {
+   ClassInit();
+
    vproto=0;
-   home=0;
    portname=0;
    hostname=0;
    user=pass=0;
@@ -60,13 +62,14 @@ void FileAccess::Init()
    file1=0;
    real_cwd=0;
    default_cwd="~";
-   cwd=xstrdup(default_cwd);
-   real_pos=-1;
+   cwd.Set(default_cwd,false,0);
+   new_cwd=0;
+   limit=FILE_END;
+   real_pos=UNKNOWN_POS;
    pos=0;
    mode=CLOSED;
    try_time=0;
    retries=0;
-   time(&event_time);
    opt_date=0;
    opt_size=0;
    error=0;
@@ -77,10 +80,10 @@ void FileAccess::Init()
    norest_manual=false;
    location=0;
    suggested_filename=0;
+   entity_content_type=0;
+   entity_charset=0;
    array_ptr=array_cnt=0;
 
-   url=0;
-
    entity_size=NO_SIZE;
    entity_date=NO_DATE;
 
@@ -99,10 +102,8 @@ FileAccess::FileAccess(const FileAccess *fa)
 {
    Init();
 
-   xfree(cwd);
-   cwd=xstrdup(fa->cwd);
-   xfree(home);
-   home=xstrdup(fa->home);
+   cwd=fa->cwd;
+   home=fa->home;
    xfree(user);
    user=xstrdup(fa->user);
    xfree(pass);
@@ -120,17 +121,16 @@ FileAccess::~FileAccess()
    xfree(vproto);
    xfree(file);
    xfree(file_url);
-   xfree(cwd);
    xfree(real_cwd);
    xfree(error);
-   xfree(home);
    xfree(user);
    xfree(pass);
    xfree(hostname);
    xfree(portname);
-   xfree(url);
    xfree(closure);
    xfree(location);
+   xfree(entity_content_type);
+   xfree(entity_charset);
    for(FileAccess **scan=&chain; *scan; scan=&((*scan)->next))
    {
       if(*scan==this)
@@ -155,59 +155,6 @@ void  FileAccess::DebugPrint(const char *prefix,const char *str,int level)
    Log::global->Write(level,buf);
 }
 
-int   FileAccess::Poll(int fd,int ev)
-{
-   struct pollfd pfd;
-   pfd.fd=fd;
-   pfd.events=ev;
-   pfd.revents=0;
-   int res=poll(&pfd,1,0);
-   if(res<1)
-      return 0;
-   if(CheckHangup(&pfd,1))
-      return -1;
-   if(pfd.revents)
-      event_time=now;
-   return pfd.revents;
-}
-
-int   FileAccess::CheckHangup(struct pollfd *pfd,int num)
-{
-   for(int i=0; i<num; i++)
-   {
-#ifdef SO_ERROR
-      char  str[256];
-      int   s_errno=0;
-      socklen_t len;
-
-      errno=0;
-
-// Where does the error number go - to errno or to the pointer?
-// It seems that to errno, but if the pointer is NULL it dumps core.
-// (solaris 2.5)
-// It seems to be different on glibc 2.0 - check both errno and s_errno
-
-      len=sizeof(s_errno);
-      getsockopt(pfd[i].fd,SOL_SOCKET,SO_ERROR,(char*)&s_errno,&len);
-      if(errno==ENOTSOCK)
-	 return 0;
-      if(errno!=0 || s_errno!=0)
-      {
-	 sprintf(str,_("Socket error (%s) - reconnecting"),
-				    strerror(errno?errno:s_errno));
-	 DebugPrint("**** ",str,0);
-	 return 1;
-      }
-#endif /* SO_ERROR */
-      if(pfd[i].revents&POLLERR)
-      {
-	 DebugPrint("**** ","POLLERR",0);
-	 return 1;
-      }
-   } /* end for */
-   return 0;
-}
-
 void FileAccess::NonBlock(int fd)
 {
    int fl=fcntl(fd,F_GETFL);
@@ -228,7 +175,7 @@ void  FileAccess::Open(const char *fn,int mode,off_t offs)
       Close();
    Resume();
    file=xstrdup(fn);
-   real_pos=-1;
+   real_pos=UNKNOWN_POS;
    pos=offs;
    this->mode=mode;
    mkdir_p=false;
@@ -240,11 +187,11 @@ void  FileAccess::Open(const char *fn,int mode,off_t offs)
    case REMOVE:
    case MAKE_DIR:
    case CHANGE_MODE:
-      LsCache::FileChanged(this,file);
+      cache->FileChanged(this,file);
       break;
    case REMOVE_DIR:
-      LsCache::FileChanged(this,file);
-      LsCache::TreeChanged(this,file);
+      cache->FileChanged(this,file);
+      cache->TreeChanged(this,file);
       break;
    default:
       break;
@@ -307,6 +254,11 @@ const char *FileAccess::StrError(int err)
       }
       return(_("Login failed"));
    case(NOT_SUPP):
+      if(error)
+      {
+   	 sprintf(str,"%s: %s",_("Operation not supported"),error);
+	 return(str);
+      }
       return(_("Operation not supported"));
    case(FILE_MOVED):
       if(error)
@@ -329,6 +281,7 @@ void FileAccess::Close()
    xfree(file); file=0;
    xfree(file_url); file_url=0;
    xfree(file1); file1=0;
+   delete new_cwd; new_cwd=0;
    mode=CLOSED;
    opt_date=0;
    opt_size=0;
@@ -339,6 +292,10 @@ void FileAccess::Close()
    ascii=false;
    norest_manual=false;
    xfree(location); location=0;
+   xfree(entity_content_type);
+   entity_content_type=0;
+   xfree(entity_charset);
+   entity_charset=0;
    ClearError();
 }
 
@@ -348,9 +305,9 @@ void FileAccess::Rename(const char *f,const char *f1)
    file1=xstrdup(f1);
    Open(f,RENAME);
 
-   LsCache::TreeChanged(this,file);
-   LsCache::FileChanged(this,file);
-   LsCache::FileChanged(this,file1);
+   cache->TreeChanged(this,file);
+   cache->FileChanged(this,file);
+   cache->FileChanged(this,file1);
 }
 
 void FileAccess::Mkdir(const char *fn,bool allp)
@@ -359,19 +316,19 @@ void FileAccess::Mkdir(const char *fn,bool allp)
    mkdir_p=allp;
 }
 
-bool FileAccess::SameLocationAs(FileAccess *fa)
+bool FileAccess::SameLocationAs(const FileAccess *fa) const
 {
-   if(strcmp(this->GetProto(),fa->GetProto()))
-      return false;
-   return true;
+   return SameSiteAs(fa);
 }
-bool FileAccess::SameSiteAs(FileAccess *fa)
+bool FileAccess::SameSiteAs(const FileAccess *fa) const
 {
-   return SameLocationAs(fa);
+   return SameProtoAs(fa);
 }
 
-const char *FileAccess::GetFileURL(const char *f,int flags)
+const char *FileAccess::GetFileURL(const char *f,int flags) const
 {
+   static char *url;
+
    const char *proto=GetVisualProto();
    if(proto[0]==0)
       return "";
@@ -387,8 +344,21 @@ const char *FileAccess::GetFileURL(const char *f,int flags)
    u.port=portname;
    if(!(flags&NO_PATH))
    {
+      if(cwd.url)
+      {
+	 Path f_path(cwd);
+	 if(f)
+	    f_path.Change(f,true);
+	 int f_path_index=url::path_index(f_path.url);
+	 xfree(url);
+	 url=u.Combine(home);
+	 url=(char*)xrealloc(url,strlen(url)+strlen(f_path.url+f_path_index)+2);
+	 strcpy(url+strlen(url),f_path.url+f_path_index);
+	 return url;
+      }
+
       if(!f || (f[0]!='/' && f[0]!='~'))
-	 f=dir_file(cwd?cwd:"~",f);
+	 f=dir_file(cwd.path?cwd.path:"~",f);
       u.path=(char*)f;
    }
 
@@ -398,7 +368,7 @@ const char *FileAccess::GetFileURL(const char *f,int flags)
    return url;
 }
 
-const char *FileAccess::GetConnectURL(int flags)
+const char *FileAccess::GetConnectURL(int flags) const
 {
    return GetFileURL(0,flags);
 }
@@ -456,8 +426,8 @@ void FileAccess::AnonymousLogin()
 
 void FileAccess::ResetLocationData()
 {
-   xfree(cwd);  cwd=xstrdup(default_cwd);
-   xfree(home); home=0;
+   cwd.Set(default_cwd,false,0);
+   home.Set((char*)0);
 }
 
 void FileAccess::SetPasswordGlobal(const char *p)
@@ -509,16 +479,15 @@ void  FileAccess::ExpandTildeInCWD()
 {
    if(home)
    {
-      if(cwd)
-	 expand_tilde(&cwd,home);
+      cwd.ExpandTilde(home);
+      if(new_cwd)
+	 new_cwd->ExpandTilde(home);
       if(real_cwd)
 	 expand_tilde(&real_cwd,home);
       if(file)
 	 expand_tilde(&file,home);
       if(file1)
 	 expand_tilde(&file1,home);
-      if(mode==CHANGE_DIR)
-	 OptimizePath(file);
    }
 }
 
@@ -558,7 +527,7 @@ int FileAccess::device_prefix_len(const char *path)
    return 0;
 }
 
-void FileAccess::OptimizePath(char *path)
+void FileAccess::Path::Optimize(char *path,int device_prefix_len)
 {
    int	 prefix_size=0;
 
@@ -577,7 +546,7 @@ void FileAccess::OptimizePath(char *path)
    else
    {
       // handle VMS and DOS devices.
-      prefix_size=device_prefix_len(path);
+      prefix_size=device_prefix_len;
    }
 
    char	 *in;
@@ -600,8 +569,8 @@ void FileAccess::OptimizePath(char *path)
    {
       if(in[0]=='/')
       {
-	 // double slash or a slash on the end
-	 if(in[1]=='/' || in[1]=='\0')
+	 // double slash
+	 if(in[1]=='/') // || (strip_trailing_slash && in[1]=='\0'))
 	 {
 	    in++;
 	    continue;
@@ -649,35 +618,27 @@ void FileAccess::OptimizePath(char *path)
 
 void FileAccess::Chdir(const char *path,bool verify)
 {
-   ExpandTildeInCWD();
+   cwd.ExpandTilde(home);
 
-   char	 *newcwd=(char*)alloca(xstrlen(cwd)+strlen(path)+2);
-
-   if(path[0]=='/')
-      strcpy(newcwd,path);
-   else if(path[0]=='~')
-      strcpy(newcwd,path);
-   else if(device_prefix_len(path)>0)
-      strcpy(newcwd,path);
-   else
-   {
-      if(cwd && cwd[0])
-      {
-	 if(last_char(cwd)=='/')
-	    sprintf(newcwd,"%s%s",cwd,path);
-	 else
-	    sprintf(newcwd,"%s/%s",cwd,path);
-      }
-      else
-	 strcpy(newcwd,path);
-   }
-
-   OptimizePath(newcwd);
+   Close();
+   new_cwd=new Path(&cwd);
+   new_cwd->Change(path,false);
 
    if(verify)
-      Open(newcwd,CHANGE_DIR);
+      Open(new_cwd->path,CHANGE_DIR);
    else
-      SetCwd(newcwd);
+   {
+      cwd.Set(new_cwd);
+      delete new_cwd;
+      new_cwd=0;
+   }
+}
+
+void FileAccess::PathVerify(const Path &p)
+{
+   Close();
+   new_cwd=new Path(p);
+   Open(new_cwd->path,CHANGE_DIR);
 }
 
 void FileAccess::Chmod(const char *file,int m)
@@ -734,6 +695,14 @@ void FileAccess::SetSuggestedFileName(const char *fn)
    suggested_filename=xstrdup(fn);
 }
 
+void FileAccess::SetFileURL(const char *u)
+{
+   xfree(file_url);
+   file_url=xstrdup(u);
+   if(new_cwd && mode==CHANGE_DIR)
+      new_cwd->SetURL(u);
+}
+
 FileAccess *SessionPool::pool[pool_size];
 
 void SessionPool::Reuse(FileAccess *f)
@@ -857,17 +826,12 @@ bool FileAccess::NotSerious(int e)
    return false;
 }
 
-void FileAccess::BumpEventTime(time_t t)
-{
-   if(event_time<t)
-      event_time=t;
-}
 void FileAccess::SetTryTime(time_t t)
 {
    try_time=t;
 }
 
-ResValue FileAccess::Query(const char *name,const char *closure)
+ResValue FileAccess::Query(const char *name,const char *closure) const
 {
    const char *prefix=res_prefix;
    if(!prefix)
@@ -877,17 +841,17 @@ ResValue FileAccess::Query(const char *name,const char *closure)
    return ResMgr::Query(fullname,closure);
 }
 
-bool FileAccess::IsBetterThan(FileAccess *fa)
+bool FileAccess::IsBetterThan(const FileAccess *fa) const
 {
    return(SameProtoAs(fa) && this->IsConnected() > fa->IsConnected());
 }
 
 void FileAccess::Reconfig(const char *) {}
-void FileAccess::ConnectVerify() {}
+void FileAccess::ConnectVerify() { mode=CONNECT_VERIFY; }
 const char *FileAccess::CurrentStatus() { return ""; }
 int FileAccess::Buffered() { return 0; }
 bool FileAccess::IOReady() { return IsOpen(); }
-int FileAccess::IsConnected() { return 0; }
+int FileAccess::IsConnected() const { return 0; }
 void FileAccess::Disconnect() {}
 void FileAccess::UseCache(bool) {}
 bool FileAccess::NeedSizeDateBeforehand() { return false; }
@@ -926,8 +890,10 @@ FileAccess *FileAccess::NextSameSite(FA *scan)
    return 0;
 }
 
-FileAccess *FileAccess::New(const char *proto,const char *host)
+FileAccess *FileAccess::New(const char *proto,const char *host,const char *port)
 {
+   ClassInit();
+
    if(!strcmp(proto,"slot"))
    {
       FA *session=ConnectionSlot::FindSession(host);
@@ -950,6 +916,9 @@ FileAccess *FileAccess::New(const char *proto,const char *host)
       }
    }
 
+   if(host)
+      session->Connect(host,port);
+
    return session;
 }
 FileAccess *FileAccess::New(const ParsedURL *u,bool dummy)
@@ -1032,13 +1001,18 @@ void FileAccessOperation::SetError(const char *e)
    error_text=xstrdup(e);
    done=true;
 }
+void FileAccessOperation::SetErrorCached(const char *e)
+{
+   xfree(error_text);
+   error_text=xasprintf(_("%s [cached]"),e);
+   done=true;
+}
 
 
 // ListInfo implementation
 ListInfo::ListInfo(FileAccess *s,const char *p)
 {
    session=s;
-   saved_cwd=0;
    result=0;
 
    exclude=0;
@@ -1050,7 +1024,7 @@ ListInfo::ListInfo(FileAccess *s,const char *p)
 
    if(session && p)
    {
-      saved_cwd=xstrdup(session->GetCwd());
+      saved_cwd=session->GetCwd();
       session->Chdir(p,false);
    }
 }
@@ -1059,10 +1033,9 @@ ListInfo::~ListInfo()
 {
    if(session)
       session->Close();
-   if(session && saved_cwd)
+   if(session && saved_cwd.path)
       session->SetCwd(saved_cwd);
    delete result;
-   xfree(saved_cwd);
 }
 
 void ListInfo::SetExclude(const char *p,PatternSet *x)
@@ -1071,6 +1044,138 @@ void ListInfo::SetExclude(const char *p,PatternSet *x)
    exclude_prefix=p;
 }
 
+// Path implementation
+void FileAccess::Path::init()
+{
+   device_prefix_len=0;
+   path=0;
+   is_file=false;
+   url=0;
+}
+FileAccess::Path::~Path()
+{
+   xfree(path);
+   xfree(url);
+}
+void FileAccess::Path::Set(const char *new_path,bool new_is_file,const char *new_url,int new_device_prefix_len)
+{
+   xfree(path);
+   path=xstrdup(new_path);
+   is_file=new_is_file;
+   xfree(url);
+   url=xstrdup(new_url);
+   device_prefix_len=new_device_prefix_len;
+}
+void FileAccess::Path::Set(const Path *o)
+{
+   Set(o->path,o->is_file,o->url,o->device_prefix_len);
+}
+void FileAccess::Path::Change(const char *new_path,bool new_is_file,const char *new_path_enc,int new_device_prefix_len)
+{
+   if(!new_path && new_path_enc)
+   {
+      char *np=alloca_strdup(new_path_enc);
+      url::decode_string(np);
+      new_path=np;
+   }
+   if(!new_path || !*new_path)
+      return;
+   const char *bn=basename_ptr(new_path);
+   if(!strcmp(bn,".") || !strcmp(bn,".."))
+      new_is_file=false;
+   if(url)
+   {
+      int path_index=url::path_index(url);
+      const char *old_url_path=url+path_index;
+      char *new_url_path=string_alloca(strlen(old_url_path)+xstrlen(new_path)*3+2);
+      strcpy(new_url_path,old_url_path);
+      if(is_file)
+      {
+	 dirname_modify(new_url_path);
+	 if(!new_url_path[0])
+	    strcpy(new_url_path,"/~");
+      }
+      if(last_char(new_url_path)!='/')
+	 strcat(new_url_path,"/");
+      if(new_path[0]!='/' && new_path[0]!='~' && new_device_prefix_len==0)
+      {
+	 if(new_path_enc)
+	    strcat(new_url_path,new_path_enc);
+	 else
+	 {
+	    char *np=alloca_strdup(new_path);
+	    Optimize(np);
+	    if(np[0]=='.' && np[1]=='.' && (np[2]=='/' || np[2]==0))
+	    {
+	       xfree(url);
+	       url=0;
+	       goto url_done;
+	    }
+	    else
+	       url::encode_string(np,new_url_path+strlen(new_url_path),URL_PATH_UNSAFE);
+	 }
+      }
+      else
+      {
+	 if(new_path_enc)
+	    strcpy(new_url_path,new_path_enc);
+	 else
+	    url::encode_string(new_path,new_url_path,URL_PATH_UNSAFE);
+      }
+      Optimize(new_url_path+(!strncmp(new_url_path,"/~",2)));
+      url=(char*)xrealloc(url,path_index+strlen(new_url_path)+1);
+      strcpy(url+path_index,new_url_path);
+   }
+url_done:
+   if(new_path[0]!='/' && new_path[0]!='~' && new_device_prefix_len==0
+   && path && path[0])
+   {
+      if(is_file)
+      {
+	 dirname_modify(path);
+	 if(!path[0])
+	    strcpy(path,"~");
+      }
+      char *newcwd=string_alloca(xstrlen(path)+xstrlen(new_path)+2);
+      if(last_char(path)=='/')
+	 sprintf(newcwd,"%s%s",path,new_path);
+      else
+	 sprintf(newcwd,"%s/%s",path,new_path);
+      new_path=newcwd;
+   }
+   xfree(path);
+   path=xstrdup(new_path);
+   device_prefix_len=new_device_prefix_len;
+   Optimize();
+   strip_trailing_slashes(path);
+   is_file=new_is_file;
+   if(!strcmp(path,"/") || !strcmp(path,"//"))
+      is_file=false;
+}
+bool FileAccess::Path::operator==(const Path &p2) const
+{
+   const Path &p1=*this;
+   if(p1.is_file!=p2.is_file)
+      return false;
+   if(xstrcmp(p1.path,p2.path))
+      return false;
+   if(xstrcmp(p1.url,p2.url))
+      return false;
+   return true;
+}
+void FileAccess::Path::ExpandTilde(const Path &home)
+{
+   if(!home.path)
+      return;
+   if(path && path[0]=='~' && (path[1]=='/' || path[1]=='\0'))
+   {
+      device_prefix_len=home.device_prefix_len;
+      if(path[1]=='\0')
+	 is_file=home.is_file;
+   }
+   expand_tilde(&path,home.path);
+}
+
 #include "modconfig.h"
 #ifndef MODULE_PROTO_FTP
 # include "ftpclass.h"
@@ -1102,11 +1207,26 @@ void ListInfo::SetExclude(const char *p,PatternSet *x)
 #else
 # define _sftp
 #endif
+bool FileAccess::class_inited;
+LsCache *FileAccess::cache;
 void FileAccess::ClassInit()
 {
+   if(class_inited)
+      return;
+   class_inited=true;
+   cache=new LsCache();
+
+   SignalHook::ClassInit();
+   ResMgr::ClassInit();
+
    _ftp;
    _file;
    _http;
    _fish;
    _sftp;
 }
+void FileAccess::ClassCleanup()
+{
+   delete cache;
+   cache=0;
+}
