@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileAccess.h,v 1.83 2005/04/04 09:59:39 lav Exp $ */
+/* $Id: FileAccess.h,v 1.92 2006/07/04 05:15:35 lav Exp $ */
 
 #ifndef FILEACCESS_H
 #define FILEACCESS_H
@@ -41,11 +41,14 @@
 #include "xmalloc.h"
 #include "ResMgr.h"
 #include "FileSet.h"
+#include "LsCache.h"
 
-#define NO_SIZE	     (-1L)
-#define NO_SIZE_YET  (-2L)
+#define NO_SIZE	     ((off_t)-1L)
+#define NO_SIZE_YET  ((off_t)-2L)
 #define NO_DATE	     ((time_t)-1L)
 #define NO_DATE_YET  ((time_t)-2L)
+#define FILE_END     ((off_t)-1L)
+#define UNKNOWN_POS  ((off_t)-1L)
 
 class ListInfo;
 class Glob;
@@ -55,7 +58,9 @@ class ArgV;
 
 class FileAccess : public SMTask
 {
+   static bool class_inited;
 public:
+   static LsCache *cache;
    enum open_mode
    {
       CLOSED,
@@ -77,14 +82,47 @@ public:
 
    struct fileinfo
    {
-      const char *file;
       off_t size;
+      const char *file;
       time_t time;
       bool get_size:1;
       bool get_time:1;
       bool is_dir:1;
    };
 
+   class Path
+   {
+      void init();
+   public:
+      int   device_prefix_len;
+      char  *path;
+      bool  is_file;
+      char  *url;
+      Path() { init(); }
+      Path(const Path *o) { init(); Set(o); }
+      Path(const Path &o) { init(); Set(o); }
+      Path(const char *new_path) { init(); Set(new_path); }
+      Path(const char *new_path,bool new_is_file,const char *new_url=0,int new_device_prefix_len=0)
+	 { init(); Set(new_path,new_is_file,new_url,new_device_prefix_len); }
+      ~Path();
+      void Set(const Path*);
+      void Set(const Path &o) { Set(&o); }
+      void Set(const char *new_path,bool new_is_file=false,const char *new_url=0,int device_prefix_len=0);
+      void SetURL(const char *u) { xfree(url); url=xstrdup(u); }
+      void Change(const char *new_path,bool new_is_file=false,const char *new_path_enc=0,int device_prefix_len=0);
+      void ExpandTilde(const Path &home);
+      static void Optimize(char *p,int dev_prefix=0);
+      void Optimize() { Optimize(path,device_prefix_len); }
+      const Path& operator=(const Path &o)
+	 {
+	    Set(&o);
+	    return *this;
+	 }
+      operator const char *() const { return path; }
+      bool operator==(const Path &p2) const;
+      bool operator!=(const Path &p2) const { return !(*this==p2); }
+   };
+
 protected:
    char	 *vproto;
    char	 *hostname;
@@ -93,30 +131,28 @@ protected:
    char  *pass;
    bool	 pass_open;
 
-   char	 *home;
    const char *default_cwd;
-   char  *cwd;
+   Path	 home;
+   Path	 cwd;
+   Path	 *new_cwd;
    char  *file;
    char	 *file_url;
    char	 *file1;
    int	 mode;
    off_t pos;
    off_t real_pos;
+   off_t limit;
 
    time_t *opt_date;
    off_t  *opt_size;
 
-   int	 Poll(int fd,int ev);
-   int   CheckHangup(struct pollfd *pfd,int num);
    static void NonBlock(int fd);
    static void CloseOnExec(int fd);
 
    void  DebugPrint(const char *prefix,const char *str,int level=9);
 
-   time_t   try_time;
-   time_t   event_time;
-   void BumpEventTime(time_t t);
-   int	 retries;
+   time_t try_time;
+   int retries;
 
    fileinfo *array_for_info;
    int	 array_ptr;
@@ -137,20 +173,17 @@ protected:
       }
    void set_home(const char *h)
       {
-	 xfree(home);
-	 home=xstrdup(h);
+	 home.Set(h);
 	 ExpandTildeInCWD();
       }
 
-   char *url;
-
    off_t  entity_size; // size of file to be sent
    time_t entity_date; // date of file to be sent
 
    char *closure;
    const char *res_prefix;
-   ResValue Query(const char *name,const char *closure=0);
-   bool QueryBool(const char *name,const char *closure=0)
+   ResValue Query(const char *name,const char *closure=0) const;
+   bool QueryBool(const char *name,const char *closure=0) const
       {
 	 return Query(name,closure).to_bool();
       }
@@ -185,24 +218,25 @@ protected:
    virtual ~FileAccess();
 
 public:
-   virtual const char *GetProto() = 0; // http, ftp, file etc
-   bool SameProtoAs(FileAccess *fa) { return !strcmp(GetProto(),fa->GetProto()); }
-   virtual FileAccess *Clone() = 0;
+   virtual const char *GetProto() const = 0; // http, ftp, file etc
+   bool SameProtoAs(const FileAccess *fa) const { return !strcmp(GetProto(),fa->GetProto()); }
+   virtual FileAccess *Clone() const = 0;
    virtual const char *ProtocolSubstitution(const char *host) { return 0; }
 
-   const char *GetVisualProto() { return vproto?vproto:GetProto(); }
+   const char *GetVisualProto() const { return vproto?vproto:GetProto(); }
    void SetVisualProto(const char *p) { xfree(vproto); vproto=xstrdup(p); }
-   const char  *GetHome() { return home; }
-   const char  *GetHostName() { return hostname; }
-   const char  *GetUser() { return user; }
-   const char  *GetPassword() { return pass; }
-   const char  *GetPort() { return portname; }
-   const char  *GetConnectURL(int flags=0);
-   const char  *GetFileURL(const char *file,int flags=0);
+   const char  *GetHome() const { return home; }
+   const char  *GetHostName() const { return hostname; }
+   const char  *GetUser() const { return user; }
+   const char  *GetPassword() const { return pass; }
+   const char  *GetPort() const { return portname; }
+   const char  *GetConnectURL(int flags=0) const;
+   const char  *GetFileURL(const char *file,int flags=0) const;
    enum { NO_PATH=1,WITH_PASSWORD=2,NO_PASSWORD=4,NO_USER=8 };
 
-   virtual void Connect(const char *h,const char *p);
-   virtual void ConnectVerify();
+   void Connect(const char *h,const char *p);
+   void ConnectVerify();
+   void PathVerify(const Path &);
    virtual void AnonymousLogin();
    virtual void Login(const char *u,const char *p);
 
@@ -210,7 +244,8 @@ public:
    virtual void ResetLocationData();
 
    virtual void Open(const char *file,int mode,off_t pos=0);
-   void SetFileURL(const char *u) { xfree(file_url); file_url=xstrdup(u); }
+   void SetFileURL(const char *u);
+   void SetLimit(off_t lim) { limit=lim; }
    void SetSize(off_t s) { entity_size=s; }
    void SetDate(time_t d) { entity_date=d; }
    void WantDate(time_t *d) { opt_date=d; }
@@ -221,8 +256,8 @@ public:
    virtual void	Rename(const char *rfile,const char *to);
    virtual void Mkdir(const char *rfile,bool allpath=false);
    virtual void Chdir(const char *dir,bool verify=true);
-   void OptimizePath(char *path);
-   void SetCwd(const char *dir) { xfree(cwd); cwd=xstrdup(dir); }
+   void ChdirAccept() { cwd=*new_cwd; }
+   void SetCwd(const Path &new_cwd) { cwd=new_cwd; }
    void Remove(const char *rfile)    { Open(rfile,REMOVE); }
    void RemoveDir(const char *dir)  { Open(dir,REMOVE_DIR); }
    void Chmod(const char *file,int m);
@@ -247,15 +282,16 @@ public:
    void SeekReal() { pos=GetRealPos(); }
    void RereadManual() { norest_manual=true; }
 
-   const char *GetCwd() { return cwd; }
-   const char *GetFile() { return file; }
+   const Path& GetCwd() const { return cwd; }
+   const Path& GetNewCwd() const { return *new_cwd; }
+   const char *GetFile() const { return file; }
 
    virtual int Do() = 0;
    virtual int Done() = 0;
 
-   virtual bool SameLocationAs(FileAccess *fa);
-   virtual bool SameSiteAs(FileAccess *fa);
-   virtual bool IsBetterThan(FileAccess *fa);
+   virtual bool SameLocationAs(const FileAccess *fa) const;
+   virtual bool SameSiteAs(const FileAccess *fa) const;
+   bool IsBetterThan(const FileAccess *fa) const;
 
    void Init();
    FileAccess() { Init(); }
@@ -267,7 +303,7 @@ public:
    bool	 IsOpen() { return !IsClosed(); }
    int	 OpenMode() { return mode; }
 
-   virtual int  IsConnected(); // level of connection (0 - not connected).
+   virtual int  IsConnected() const; // level of connection (0 - not connected).
    virtual void Disconnect();
    virtual void UseCache(bool);
    virtual bool NeedSizeDateBeforehand();
@@ -299,7 +335,7 @@ public:
    virtual ListInfo *MakeListInfo(const char *path=0);
    virtual Glob *MakeGlob(const char *pattern);
    virtual DirList *MakeDirList(ArgV *a);
-   virtual FileSet *ParseLongList(const char *buf,int len,int *err=0) { return 0; }
+   virtual FileSet *ParseLongList(const char *buf,int len,int *err=0) const { return 0; }
 
    static bool NotSerious(int err);
 
@@ -330,7 +366,7 @@ public:
 	 (void)new Protocol(proto,creator);
       }
 
-   static FileAccess *New(const char *proto,const char *host=0);
+   static FileAccess *New(const char *proto,const char *host=0,const char *port=0);
    static FileAccess *New(const class ParsedURL *u,bool dummy=true);
 
    void SetPasswordGlobal(const char *p);
@@ -356,6 +392,7 @@ public:
    const char *GetLogContext() { return hostname; }
 
    static void ClassInit();
+   static void ClassCleanup();
 };
 
 class FileAccessOperation : public SMTask
@@ -391,7 +428,7 @@ class ListInfo : public FileAccessOperation
 {
 protected:
    FileAccess *session;
-   char *saved_cwd;
+   FileAccess::Path saved_cwd;
    FileSet *result;
 
    const char *exclude_prefix;
