@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1999-2001 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.cc,v 1.204 2005/12/31 07:01:36 lav Exp $ */
+/* $Id: Http.cc,v 1.214 2006/07/04 05:15:36 lav Exp $ */
 
 #include <config.h>
 #include "trio.h"
@@ -98,6 +98,8 @@ void Http::Init()
    no_cache_this=false;
    no_cache=false;
 
+   use_propfind_now=true;
+
    hftp=false;
    https=false;
    use_head=true;
@@ -139,7 +141,7 @@ void Http::MoveConnectionHere(Http *o)
    sock=o->sock; o->sock=-1;
    rate_limit=o->rate_limit; o->rate_limit=0;
    last_method=o->last_method; o->last_method=0;
-   BumpEventTime(o->event_time);
+   timeout_timer.Reset(o->timeout_timer);
    state=CONNECTED;
    o->Disconnect();
    ResumeInternal();
@@ -233,6 +235,7 @@ void Http::Close()
    array_send=0;
    no_cache_this=false;
    no_ranges=false;
+   use_propfind_now=QueryBool("use-propfind",hostname);
    special=HTTP_NONE;
    xfree(special_data);
    special_data=0;
@@ -574,7 +577,12 @@ void Http::SendRequest(const char *connection,const char *f)
    retrieve:
       SendMethod("GET",efile);
       if(pos>0 && !no_ranges)
-	 Send("Range: bytes=%lld-\r\n",(long long)pos);
+      {
+	 if(limit==FILE_END)
+	    Send("Range: bytes=%lld-\r\n",(long long)pos);
+	 else
+	    Send("Range: bytes=%lld-%lld\r\n",(long long)pos,(long long)limit-1);
+      }
       break;
 
    case STORE:
@@ -589,10 +597,18 @@ void Http::SendRequest(const char *connection,const char *f)
       if(entity_size>=0)
 	 Send("Content-length: %lld\r\n",(long long)(entity_size-pos));
       if(pos>0 && entity_size<0)
-	 Send("Range: bytes=%lld-\r\n",(long long)pos);
+      {
+	 if(limit==FILE_END)
+	    Send("Range: bytes=%lld-\r\n",(long long)pos);
+	 else
+	    Send("Range: bytes=%lld-%lld\r\n",(long long)pos,(long long)limit-1);
+      }
       else if(pos>0)
+      {
 	 Send("Range: bytes=%lld-%lld/%lld\r\n",(long long)pos,
-		     (long long)(entity_size-1),(long long)entity_size);
+		     (long long)((limit==FILE_END || limit>entity_size ? entity_size : limit)-1),
+		     (long long)entity_size);
+      }
       if(entity_date!=NO_DATE)
       {
 	 char d[256];
@@ -614,7 +630,17 @@ void Http::SendRequest(const char *connection,const char *f)
       if(last_char(efile)!='/' && add_slash)
 	 strcat(efile,"/");
       if(mode==CHANGE_DIR)
-	 SendMethod("HEAD",efile);
+      {
+	 if(use_propfind_now)
+	 {
+	    SendMethod("PROPFIND",efile);
+	    Send("Depth: 0\r\n"); // no directory listing required
+	    Send("Content-Type: text/xml\r\n");
+	    Send("Content-Length: %d\r\n",strlen(allprop));
+	 }
+	 else
+	    SendMethod("HEAD",efile);
+      }
       else if(mode==LONG_LIST)
 	 goto retrieve;
       else if(mode==MAKE_DIR)
@@ -672,7 +698,7 @@ void Http::SendRequest(const char *connection,const char *f)
 	 Send("%s",special_data);
       entity_size=NO_SIZE;
    }
-   else if(mode==MP_LIST)
+   else if(mode==MP_LIST || (mode==CHANGE_DIR && use_propfind_now))
    {
       Send("%s",allprop);
    }
@@ -719,6 +745,7 @@ static const char *extract_quoted_header_value(const char *value)
 	    value++;
 	 *store++=*value++;
       }
+      *store=0;
    }
    else
    {
@@ -736,6 +763,8 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       long long bs=0;
       if(1!=sscanf(value,"%lld",&bs))
 	 return;
+      if(bs<0) // try to workaround broken servers
+	 bs+=0x100000000LL;
       body_size=bs;
       if(pos==0 && mode!=STORE && mode!=MAKE_DIR)
 	 entity_size=body_size;
@@ -930,15 +959,11 @@ int Http::Do()
    int len;
 
    // check if idle time exceeded
-   if(mode==CLOSED && sock!=-1 && idle>0)
+   if(mode==CLOSED && sock!=-1 && idle_timer.Stopped())
    {
-      if(now >= idle_start+idle)
-      {
-	 DebugPrint("---- ",_("Closing idle connection"),1);
-	 Disconnect();
-	 return m;
-      }
-      TimeoutS(idle_start+idle-time_t(now));
+      DebugPrint("---- ",_("Closing idle connection"),1);
+      Disconnect();
+      return m;
    }
 
    if(home.path==0)
@@ -1076,8 +1101,8 @@ int Http::Do()
 	 return MOVED;
       }
       KeepAlive(sock);
-      SetSocketBuffer(sock,socket_buffer);
-      SetSocketMaxseg(sock,socket_maxseg);
+      SetSocketBuffer(sock);
+      SetSocketMaxseg(sock);
       NonBlock(sock);
       CloseOnExec(sock);
 
@@ -1098,7 +1123,7 @@ int Http::Do()
       }
       state=CONNECTING;
       m=MOVED;
-      event_time=now;
+      timeout_timer.Reset();
 
    case CONNECTING:
       res=Poll(sock,POLLOUT);
@@ -1213,8 +1238,8 @@ int Http::Do()
 	 Disconnect();
 	 return MOVED;
       }
-      BumpEventTime(send_buf->EventTime());
-      BumpEventTime(recv_buf->EventTime());
+      timeout_timer.Reset(send_buf->EventTime());
+      timeout_timer.Reset(recv_buf->EventTime());
       if(CheckTimeout())
 	 return MOVED;
       recv_buf->Get(&buf,&len);
@@ -1490,6 +1515,13 @@ int Http::Do()
 		  if(!xstrcmp(last_method,"MKCOL"))
 		     ResMgr::Set("http:use-mkcol",hostname,"no");
 	       }
+	       if(mode==CHANGE_DIR && !xstrcmp(last_method,"PROPFIND"))
+	       {
+		  use_propfind_now=false;
+		  try_time=0;
+		  Disconnect();
+		  return MOVED;
+	       }
 	       code=NOT_SUPP;
 	       closure=last_method;
 	    }
@@ -1546,8 +1578,8 @@ int Http::Do()
 	    if(recv_buf->Size()>0 || (recv_buf->Size()==0 && recv_buf->Eof()))
 	       m=MOVED;
 	 }
-	 BumpEventTime(send_buf->EventTime());
-	 BumpEventTime(recv_buf->EventTime());
+	 timeout_timer.Reset(send_buf->EventTime());
+	 timeout_timer.Reset(recv_buf->EventTime());
 	 if(recv_buf->Size()==0)
 	 {
 	    // check if ranges were emulated by squid
@@ -1639,15 +1671,18 @@ int Http::Read(void *buf,int size)
 	 }
 	 return 0;
       }
-      if(body_size>=0 && bytes_received>=body_size)
+      if(!chunked)
       {
-	 DebugPrint("---- ",_("Received all"));
-	 return 0; // all received
-      }
-      if(entity_size>=0 && pos>=entity_size)
-      {
-	 DebugPrint("---- ",_("Received all (total)"));
-	 return 0;
+	 if(body_size>=0 && bytes_received>=body_size)
+	 {
+	    DebugPrint("---- ",_("Received all"));
+	    return 0; // all received
+	 }
+	 if(entity_size>=0 && pos>=entity_size)
+	 {
+	    DebugPrint("---- ",_("Received all (total)"));
+	    return 0;
+	 }
       }
       if(size1==0)
 	 return DO_AGAIN;
@@ -1678,6 +1713,7 @@ int Http::Read(void *buf,int size)
 	 }
 	 if(chunk_size==0) // eof
 	 {
+	    DebugPrint("---- ",_("Received last chunk"));
 	    // headers may follow
 	    chunked_trailer=true;
 	    state=RECEIVING_HEADER;
@@ -1903,14 +1939,14 @@ void Http::Reconfig(const char *name)
    }
 
    if(sock!=-1)
-      SetSocketBuffer(sock,socket_buffer);
+      SetSocketBuffer(sock);
    if(proxy && proxy_port==0)
       proxy_port=xstrdup(HTTP_DEFAULT_PROXY_PORT);
 
    user_agent=ResMgr::Query("http:user-agent",c);
 }
 
-bool Http::SameSiteAs(FileAccess *fa)
+bool Http::SameSiteAs(const FileAccess *fa) const
 {
    if(!SameProtoAs(fa))
       return false;
@@ -1919,7 +1955,7 @@ bool Http::SameSiteAs(FileAccess *fa)
    && !xstrcmp(user,o->user) && !xstrcmp(pass,o->pass));
 }
 
-bool Http::SameLocationAs(FileAccess *fa)
+bool Http::SameLocationAs(const FileAccess *fa) const
 {
    if(!SameSiteAs(fa))
       return false;
@@ -1934,6 +1970,7 @@ void Http::ResetLocationData()
    super::ResetLocationData();
    Reconfig();
    state=DISCONNECTED;
+   use_propfind_now=QueryBool("use-propfind",hostname);
 }
 
 DirList *Http::MakeDirList(ArgV *args)
