@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1998 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Job.cc,v 1.33 2005/10/18 07:54:52 lav Exp $ */
+/* $Id: Job.cc,v 1.38 2006/06/13 14:35:37 lav Exp $ */
 
 #include <config.h>
 #include <stdlib.h>
@@ -193,12 +193,13 @@ void Job::Kill(Job *j)
       // we have to simulate normal death...
       Job *r=new KilledJob();
       r->parent=j->parent;
-      r->cmdline=j->cmdline;
-      j->cmdline=0;
+      r->cmdline=j->cmdline,j->cmdline=0;
+      r->waiting=j->waiting,j->waiting=0;
+      r->waiting_num=j->waiting_num,j->waiting_num=0;
       j->parent->ReplaceWaiting(j,r);
    }
    assert(FindWhoWaitsFor(j)==0);
-   Delete(j);
+   j->DeleteLater();
 }
 
 void Job::Kill(int n)
@@ -210,13 +211,16 @@ void Job::Kill(int n)
 
 void Job::KillAll()
 {
-   for(Job *scan=chain; scan; )
-   {
-      Job *tmp=scan;
-      scan=scan->next;
-      if(tmp->jobno>=0)
-	 Job::Kill(tmp);
-   }
+   for(Job *scan=chain; scan; scan=scan->next)
+      if(scan->jobno>=0)
+	 Job::Kill(scan);
+   CollectGarbage();
+}
+void Job::Cleanup()
+{
+   for(Job *scan=chain; scan; scan=scan->next)
+      Job::Kill(scan);
+   CollectGarbage();
 }
 
 void  Job::SendSig(int n,int sig)
@@ -335,7 +339,7 @@ void  Job::ListDoneJobs()
    for(Job *scan=chain; scan; scan=scan->next)
    {
       if(scan->jobno>=0 && (scan->parent==this || scan->parent==0)
-         && scan->Done())
+         && !scan->deleting && scan->Done())
       {
 	 fprintf(f,_("[%d] Done (%s)"),scan->jobno,
 	    scan->cmdline?scan->cmdline:"?");
@@ -354,7 +358,7 @@ void  Job::BuryDoneJobs()
    for(Job *scan=chain; scan; scan=scan->next)
    {
       if((scan->parent==this || scan->parent==0) && scan->jobno>=0
-		  && scan->Done())
+		  && !scan->deleting && scan->Done())
 	 scan->deleting=true;
    }
    CollectGarbage();
@@ -385,11 +389,9 @@ void Job::eprintf(const char *fmt,...)
 }
 
 
-FileAccess *SessionJob::Clone()
+FileAccess *SessionJob::Clone() const
 {
-   FileAccess *tmp=session;
-   session=session->Clone();
-   return tmp;
+   return session->Clone();
 }
 
 SessionJob::SessionJob(FileAccess *f)
@@ -546,3 +548,14 @@ bool Job::CheckForWaitLoop(Job *parent)
 	 return true;
    return false;
 }
+
+void Job::WaitDone()
+{
+   for(;;)
+   {
+      SMTask::Schedule();
+      if(Done())
+	 break;
+      SMTask::Block();
+   }
+}
