@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2002 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,6 +41,7 @@ class SMTask
 
 protected:
    int	 running;
+   int	 ref_count;
    bool	 deleting;
 
    enum
@@ -89,8 +90,17 @@ public:
 
    void DeleteLater() { deleting=true; }
    static void Delete(SMTask *);
+   static SMTask *_MakeRef(SMTask *task) { if(task) task->ref_count++; return task; }
+   static void _DeleteRef(SMTask *task)  { if(task) task->ref_count--; Delete(task); }
+#if __GNUC__
+   template<typename T> static void DeleteRef(T *&task) { _DeleteRef(task); task=0; }
+   template<typename T> static T *MakeRef(T *task) { _MakeRef(task); return task; }
+#else
+# define DeleteRef(task) do { _DeleteRef((task)); (task)=0; } while(0)
+# define MakeRef(task)   (_MakeRef((task)),(task))
+#endif
    static int Roll(SMTask *);
-   static void RollAll(int max_time);
+   static void RollAll(const TimeInterval &max_time);
 
    static SMTask *current;
 
@@ -100,6 +110,7 @@ public:
    void Leave() { Leave(this); }
 
    static int TaskCount();
+   static void PrintTasks();
    static bool NonFatalError(int err);
    static bool TemporaryNetworkError(int err);
 
