@@ -18,10 +18,11 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Speedometer.cc,v 1.16 2002/08/09 12:25:35 lav Exp $ */
+/* $Id: Speedometer.cc,v 1.20 2006/05/24 07:30:25 lav Exp $ */
 
 #include <config.h>
 #include <math.h>
+#include <stdlib.h>
 #include "Speedometer.h"
 #include "misc.h"
 #include "xstring.h"
@@ -46,7 +47,7 @@ Speedometer::Speedometer(const char *p)
 }
 bool Speedometer::Valid()
 {
-   return now>=start+1 && now<last_bytes+period;
+   return now>=start+TimeDiff(1,0) && now<last_bytes+TimeDiff(period,0);
 }
 int Speedometer::Do()
 {
@@ -61,6 +62,9 @@ float Speedometer::Get()
 }
 void Speedometer::Add(int b)
 {
+   if(b==0 && TimeDiff(now,last_second).MilliSeconds()<100)
+      return;
+
    // This makes Speedometer start only when first data come.
    if(rate==0)
       Reset();
@@ -123,14 +127,14 @@ const char *Speedometer::GetETAStrFromTime(long eta)
    long eta2=0;
    long ueta=0;
    long ueta2=0;
-   char letter=0;
-   char letter2=0;
+   const char *letter=0;
+   const char *letter2=0;
 
    // for translator: only first letter matters
-   const char day_c=_("day")[0];
-   const char hour_c=_("hour")[0];
-   const char minute_c=_("minute")[0];
-   const char second_c=_("second")[0];
+   const char *day_c=_("day");
+   const char *hour_c=_("hour");
+   const char *minute_c=_("minute");
+   const char *second_c=_("second");
 
    // for translator: Estimated Time of Arrival.
    const char *tr_eta=_("eta:");
@@ -145,8 +149,8 @@ const char *Speedometer::GetETAStrFromTime(long eta)
 	 if(ueta<10)
 	 {
 	    letter2=hour_c;
-	    ueta2=((eta2<HOUR/2?eta2+DAY:eta2)+HOUR/2)/HOUR;
-	    if(ueta2>0 && eta2<0)
+	    ueta2=((eta2<-HOUR/2?eta2+DAY:eta2)+HOUR/2)/HOUR;
+	    if(ueta2>0 && eta2<-HOUR/2)
 	       ueta--;
 	 }
       }
@@ -158,8 +162,8 @@ const char *Speedometer::GetETAStrFromTime(long eta)
 	 if(ueta<10)
 	 {
 	    letter2=minute_c;
-	    ueta2=((eta2<MINUTE/2?eta2+HOUR:eta2)+MINUTE/2)/MINUTE;
-	    if(ueta2>0 && eta2<0)
+	    ueta2=((eta2<-MINUTE/2?eta2+HOUR:eta2)+MINUTE/2)/MINUTE;
+	    if(ueta2>0 && eta2<-MINUTE/2)
 	       ueta--;
 	 }
       }
@@ -174,9 +178,10 @@ const char *Speedometer::GetETAStrFromTime(long eta)
 	 letter=second_c;
       }
       if(letter2 && ueta2>0)
-	 sprintf(buf_eta,"%s%ld%c%ld%c",tr_eta,ueta,letter,ueta2,letter2);
+	 sprintf(buf_eta,"%s%ld%.*s%ld%.*s",tr_eta,ueta,mblen(letter,strlen(letter)),letter,
+						   ueta2,mblen(letter2,strlen(letter2)),letter2);
       else
-	 sprintf(buf_eta,"%s%ld%c",tr_eta,ueta,letter);
+	 sprintf(buf_eta,"%s%ld%.*s",tr_eta,ueta,mblen(letter,strlen(letter)),letter);
    }
    else // verbose eta (by Ben Winslow)
    {
@@ -186,20 +191,20 @@ const char *Speedometer::GetETAStrFromTime(long eta)
       if(eta>=DAY)
       {
 	 unit=eta/DAY;
-	 sprintf(buf_eta+strlen(buf_eta), "%ld%c", unit, day_c);
+	 sprintf(buf_eta+strlen(buf_eta),"%ld%.*s",unit,mblen(day_c,strlen(day_c)),day_c);
       }
       if(eta>=HOUR)
       {
 	 unit=(eta/HOUR)%24;
-	 sprintf(buf_eta+strlen(buf_eta), "%ld%c", unit, hour_c);
+	 sprintf(buf_eta+strlen(buf_eta),"%ld%.*s",unit,mblen(hour_c,strlen(hour_c)),hour_c);
       }
       if(eta>=MINUTE)
       {
 	 unit=(eta/MINUTE)%60;
-	 sprintf(buf_eta+strlen(buf_eta), "%ld%c", unit, minute_c);
+	 sprintf(buf_eta+strlen(buf_eta),"%ld%.*s",unit,mblen(minute_c,strlen(minute_c)),minute_c);
       }
       unit=eta%60;
-      sprintf(buf_eta+strlen(buf_eta), "%ld%c", unit, second_c);
+      sprintf(buf_eta+strlen(buf_eta), "%ld%.*s",unit,mblen(second_c,strlen(second_c)),second_c);
    }
    return buf_eta;
 }
