@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2005 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: commands.cc,v 1.236 2005/12/27 06:25:22 lav Exp $ */
+/* $Id: commands.cc,v 1.254 2006/06/30 14:10:29 lav Exp $ */
 
 #include <config.h>
 
@@ -36,6 +36,7 @@
 #ifdef HAVE_DLFCN_H
 # include <dlfcn.h>
 #endif
+#include <mbswidth.h>
 
 #include "CmdExec.h"
 #include "GetJob.h"
@@ -65,7 +66,6 @@
 #include "bookmark.h"
 #include "log.h"
 #include "module.h"
-#include "getopt.h"
 #include "FileCopy.h"
 #include "DummyProto.h"
 #include "QueueFeeder.h"
@@ -95,7 +95,7 @@ CMD(ver);   CMD(close);  CMD(bookmark);CMD(lftp);
 CMD(echo);  CMD(suspend);CMD(sleep);   CMD(slot);
 CMD(at);    CMD(find);   CMD(command); CMD(module);
 CMD(lpwd);  CMD(glob);	 CMD(chmod);   CMD(queue);
-CMD(repeat);CMD(get1);   CMD(history); CMD(tasks);
+CMD(repeat);CMD(get1);   CMD(tasks);
 
 #ifdef MODULE_CMD_MIRROR
 # define cmd_mirror 0
@@ -190,6 +190,7 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	    " -S                   - sort by file size\n"
 	    " --user, --group, --perms, --date, --linkcount, --links\n"
 	    "                      - show individual fields\n"
+	    " --time-style=STYLE   - use specified time format\n"
 	    "\n"
 	    "By default, cls output is cached, to see new listing use `recls' or\n"
 	    "`cache flush'.\n"
@@ -264,13 +265,6 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 " -a  all types\n")},
    {"help",    cmd_help,   N_("help [<cmd>]"),
 	 N_("Print help for command <cmd>, or list of available commands\n")},
-   {"history", cmd_history,N_("history -w file|-r file|-c|-l [cnt]"),
-	 N_(" -w <file> Write history to file.\n"
-	 " -r <file> Read history from file; appends to current history.\n"
-	 " -c  Clear the history.\n"
-	 " -l  List the history (default).\n"
-	 "Optional argument cnt specifies the number of history lines to list,\n"
-	 "or \"all\" to list all entries.\n")},
    {"jobs",    cmd_jobs,   "jobs [-v]",
 	 N_("List running jobs. -v means verbose, several -v can be specified.\n")},
    {"kill",    cmd_kill,   N_("kill all|<job_no>"),
@@ -421,9 +415,11 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
    {"renlist", cmd_ls,	    0,
 	 N_("Usage: renlist [<args>]\n"
 	 "Same as `nlist', but don't look in cache\n")},
-   {"repeat",  cmd_repeat, N_("repeat [delay] [command]"),
+   {"repeat",  cmd_repeat, N_("repeat [OPTS] [delay] [command]"),
 	 N_("Repeat specified command with a delay between iterations.\n"
-	 "Default delay is one second, default command is empty.\n")},
+	 "Default delay is one second, default command is empty.\n"
+	 " -c <count>  number of iterations\n"
+	 " -d <delay>  delay between iterations.\n")},
    {"reput",   cmd_get,    0,
 	 N_("Usage: reput <lfile> [-o <rfile>]\n"
 	 "Same as `put -c'\n")},
@@ -619,7 +615,7 @@ Job *CmdExec::builtin_cd()
    if(url)
       new_cwd.SetURL(url);
    if(!verify_path || background
-   || (!verify_path_cached && LsCache::IsDirectory(session,dir)==1))
+   || (!verify_path_cached && FileAccess::cache->IsDirectory(session,dir)==1))
    {
       cwd_history.Set(session,old_cwd);
       session->SetCwd(new_cwd);
@@ -637,12 +633,25 @@ Job *CmdExec::builtin_cd()
 Job *CmdExec::builtin_exit()
 {
    bool bg=false;
+   bool kill=false;
    int code=prev_exit_code;
-   if(args->count()>=2)
+   CmdExec *exec=this;
+   const char *a;
+   args->rewind();
+   while((a=args->getnext())!=0)
    {
-      const char *a=args->getarg(1);
       if(!strcmp(a,"bg"))
 	 bg=true;
+      if(!strcmp(a,"top") || !strcmp(a,"bg"))
+      {
+	 if(top)
+	    exec=top;
+      }
+      else if(!strcmp(a,"kill"))
+      {
+	 kill=true;
+	 bg=false;
+      }
       else if(sscanf(a,"%i",&code)!=1)
       {
 	 eprintf(_("Usage: %s [<exit_code>]\n"),args->a0());
@@ -650,7 +659,7 @@ Job *CmdExec::builtin_exit()
       }
    }
    // Note: one job is this CmdExec.
-   if(!bg && top_level
+   if(!bg && exec->top_level
    && !ResMgr::QueryBool("cmd:move-background",0) && NumberOfJobs()>1)
    {
       eprintf(_(
@@ -659,11 +668,26 @@ Job *CmdExec::builtin_exit()
       ));
       return 0;
    }
-   while(!Done())
-      RemoveFeeder();
-   exit_code=prev_exit_code=code;
+   if(kill)
+      Job::KillAll();
+   exec->Exit(code);
+   exit_code=code;
    return 0;
 }
+void CmdExec::Exit(int code)
+{
+   while(feeder)
+      RemoveFeeder();
+   if(interactive)
+   {
+      ListDoneJobs();
+      BuryDoneJobs();
+      if(FindJob(last_bg)==0)
+	 last_bg=-1;
+   }
+   exit_code=prev_exit_code=code;
+   return;
+}
 
 CmdFeeder *lftp_feeder=0;
 Job *CmdExec::builtin_lftp()
@@ -885,18 +909,6 @@ Job *CmdExec::builtin_open()
 	 {
 	    cwd_history.Set(session,session->GetCwd());
 
-	    FileAccess *new_session=FileAccess::New(uc.proto,uc.host);
-	    if(!new_session)
-	    {
-	       eprintf("%s: %s%s\n",args->a0(),uc.proto,
-			_(" - not supported protocol"));
-	       return 0;
-	    }
-
-	    saved_session=session;
-	    session=0;
-	    ChangeSession(new_session);
-
 	    if(uc.user && !user)
 	       user=uc.user;
 	    if(uc.pass && !pass)
@@ -909,6 +921,18 @@ Job *CmdExec::builtin_open()
 	       port=uc.port;
 	    if(uc.path && !path)
 	       path=uc.path;
+
+	    FileAccess *new_session=FileAccess::New(uc.proto,host,port);
+	    if(!new_session)
+	    {
+	       eprintf("%s: %s%s\n",args->a0(),uc.proto,
+			_(" - not supported protocol"));
+	       return 0;
+	    }
+
+	    saved_session=session;
+	    session=0;
+	    ChangeSession(new_session);
 	 }
 
 	 // user gets substituted only if no proto is specified.
@@ -1365,7 +1389,6 @@ CMD(ls)
    FileCopy *c=FileCopy::New(src_peer,dst_peer,false);
    c->DontCopyDate();
    c->LineBuffered();
-   c->Ascii();
 
    CopyJob *j=new CopyJob(c,a,op);
    if(!output || output->usesfd(1))
@@ -1395,6 +1418,7 @@ const char *FileSetOutput::parse_argv(ArgV *a)
       {"nocase", no_argument,0,'i'},
       {"sortnocase", no_argument,0,'I'},
       {"dirsfirst", no_argument,0,'D'},
+      {"time-style", required_argument,0,256+'T'},
 
       {"sort", required_argument,0, 0},
       {"reverse",no_argument,0,'r'},
@@ -1407,6 +1431,8 @@ const char *FileSetOutput::parse_argv(ArgV *a)
       {0,0,0,0}
    };
 
+   const char *time_style=ResMgr::Query("cmd:time-style",0);
+
    int opt, longopt;
    while((opt=a->getopt_long(":1BdFhiklqsDISrt", cls_options, &longopt))!=EOF)
    {
@@ -1485,6 +1511,9 @@ const char *FileSetOutput::parse_argv(ArgV *a)
       case('r'):
 	 sort_reverse = true;
 	 break;
+      case(256+'T'):
+	 time_style=optarg;
+	 break;
 
       default:
 	 /* silly getopt won't give us its error instead of printing it, oh well.
@@ -1498,6 +1527,35 @@ const char *FileSetOutput::parse_argv(ArgV *a)
       a->delarg(1);
    a->rewind();
 
+   xfree(time_fmt); time_fmt=0;
+   if(time_style && time_style[0]) {
+      if(time_style[0]=='+')
+	 time_fmt=xstrdup(time_style+1);
+      else if(!strcmp(time_style,"full-iso"))
+//	 time_fmt=xstrdup("%Y-%m-%d %H:%M:%S.%N %z"); // %N and %z are GNU extensions
+	 time_fmt=xstrdup("%Y-%m-%d %H:%M:%S");
+      else if(!strcmp(time_style,"long-iso"))
+	 time_fmt=xstrdup("%Y-%m-%d %H:%M");
+      else if(!strcmp(time_style,"iso"))
+	 time_fmt=xstrdup("%Y-%m-%d \n%m-%d %H:%M");
+      else
+	 time_fmt=xstrdup(time_style);
+      need_exact_time=false;
+      if(time_fmt) {
+	 static const char exact_fmts[][3]={"%H","%M","%S","%N",""};
+	 int sep=strcspn(time_fmt,"\n|");
+	 for(int i=0; exact_fmts[i][0]; i++) {
+	    const char *f=strstr(time_fmt,exact_fmts[i]);
+	    if(!f)
+	       continue;
+	    if(i>1 || sep>f-time_fmt) {
+	       need_exact_time=true;
+	       break;
+	    }
+	 }
+      }
+   }
+
    return NULL;
 }
 
@@ -1605,7 +1663,7 @@ CMD(get)
    }
    if(!strcmp(op,"pget"))
    {
-      opts="+n:euO:";
+      opts="+n:ceuO:";
       n_conn=-1;
    }
    else if(!strcmp(op,"put") || !strcmp(op,"reput"))
@@ -1729,7 +1787,7 @@ CMD(get)
    }
    else
    {
-      pgetJob *j=new pgetJob(session->Clone(),get_args);
+      pgetJob *j=new pgetJob(session->Clone(),get_args,cont);
       if(n_conn!=-1)
 	 j->SetMaxConn(n_conn);
       if(del_target)
@@ -2184,7 +2242,7 @@ CMD(set)
       closure=sl+1;
    }
 
-   ResDecl *type;
+   const ResType *type;
    // find type of given variable
    const char *msg=ResMgr::FindVar(a,&type);
    if(msg)
@@ -2344,9 +2402,9 @@ CMD(cache)  // cache control
 
    exit_code=0;
    if(!op || !strcasecmp(op,"status"))
-      LsCache::List();
+      FileAccess::cache->List();
    else if(!strcasecmp(op,"flush"))
-      LsCache::Flush();
+      FileAccess::cache->Flush();
    else if(!strcasecmp(op,"on"))
       ResMgr::Set("cache:enable",0,"yes");
    else if(!strcasecmp(op,"off"))
@@ -2498,7 +2556,7 @@ CMD(help)
 CMD(ver)
 {
    printf(
-      _("LFTP | Version %s | Copyright (c) 1996-2005 Alexander V. Lukyanov\n"),VERSION);
+      _("LFTP | Version %s | Copyright (c) 1996-%d Alexander V. Lukyanov\n"),VERSION,2006);
    printf("\n");
    printf(
       _("LFTP is free software, covered by the GNU General Public License, and you are\n"
@@ -2509,8 +2567,14 @@ CMD(ver)
       _("Send bug reports and questions to <%s>.\n"),"lftp@uniyar.ac.ru");
 
 #if defined(HAVE_DLOPEN) && defined(RTLD_DEFAULT)
+   /* Show some of loaded libraries. Modules can load those libraries on
+      demand so use dlsym to avoid linking with them just for showing version. */
    printf("\n");
-   printf(_("Libraries used: "));
+   const char *msg=_("Libraries used: ");
+   int mbflags=MBSW_ACCEPT_INVALID|MBSW_ACCEPT_UNPRINTABLE;
+   int col=mbswidth(msg,mbflags);
+   int width=parent->status_line?parent->status_line->GetWidth():80;
+   printf("%s",msg);
 
    struct VersionInfo
    {
@@ -2554,6 +2618,7 @@ CMD(ver)
       {"OpenSSL",    "SSL_version_str",	     VersionInfo::STRING_PTR,"OpenSSL "},
       {"GnuTLS",     "gnutls_check_version", VersionInfo::FUNC0,     0},
       {"libiconv",   "_libiconv_version",    VersionInfo::INT8_8,    0},
+      {"zlib",	     "zlibVersion",	     VersionInfo::FUNC0,     0},
       {0}
    };
 
@@ -2563,7 +2628,19 @@ CMD(ver)
       const char *v=scan->query();
       if(!v)
 	 continue;
-      printf("%s%s %s",need_comma?", ":"",scan->lib_name,v);
+      char buf[256];
+      sprintf(buf,", %s %s",scan->lib_name,v);
+      int skip=need_comma?0:2;
+      int w=mbswidth(buf+skip,mbflags);
+      if(col+w>=width)
+      {
+	 buf[1]='\n';
+	 col=w-2+skip;
+	 skip/=2;
+      }
+      else
+	 col+=w;
+      printf("%s",buf+skip);
       need_comma=true;
    }
    printf("\n");
@@ -3137,9 +3214,11 @@ CMD(get1)
    if(args->getnext()!=0)
       goto usage;
 
+   bool auto_rename=false;
    if(dst==0 || dst[0]==0)
    {
       dst=basename_ptr(src);
+      auto_rename=true;
    }
    else
    {
@@ -3152,6 +3231,7 @@ CMD(get1)
 	    strcpy(dst1,dst);
 	    strcat(dst1,bn);
 	    dst=dst1;
+	    auto_rename=true;
 	 }
       }
    }
@@ -3192,6 +3272,7 @@ CMD(get1)
       dst_peer=FileCopyPeerFDStream::NewPut(dst,cont||target_region_begin>0);
    else
       dst_peer=new FileCopyPeerFA(&dst_url,FA::STORE);
+   dst_peer->AutoRename(auto_rename);
    if(!cont && (target_region_begin>0 || target_region_end!=FILE_END))
       dst_peer->SetRange(target_region_begin,target_region_end);
 
@@ -3203,82 +3284,6 @@ CMD(get1)
    return new CopyJob(c,src,args->a0());
 }
 
-CMD(history)
-{
-   enum { READ, WRITE, CLEAR, LIST } mode = LIST;
-   const char *fn = NULL;
-   static struct option history_options[]=
-   {
-      {"read",required_argument,0,'r'},
-      {"write",required_argument,0,'w'},
-      {"clear",no_argument,0,'c'},
-      {"list",required_argument,0,'l'},
-      {0,0,0,0}
-   };
-
-   exit_code=0;
-   int opt;
-   while((opt=args->getopt_long("+r:w:cl",history_options,0))!=EOF) {
-      switch(opt) {
-      case 'r':
-	 mode = READ;
-	 fn = optarg;
-	 break;
-      case 'w':
-	 mode = WRITE;
-	 fn = optarg;
-	 break;
-      case 'c':
-	 mode = CLEAR;
-	 break;
-      case 'l':
-	 mode = LIST;
-	 break;
-      case '?':
-	 eprintf(_("Try `help %s' for more information.\n"),args->a0());
-	 return 0;
-      }
-   }
-
-   int cnt = 16;
-   if(const char *arg = args->getcurr()) {
-      if(!strcasecmp(arg, "all"))
-	 cnt = -1;
-      else if(isdigit(arg[0]))
-	 cnt = atoi(arg);
-      else {
-	 eprintf(_("%s: %s - not a number\n"), args->a0(), args->getcurr());
-	 exit_code=1;
-	 return 0;
-      }
-   }
-
-   switch(mode) {
-   case READ:
-      if(int err = lftp_history_read(fn)) {
-	 eprintf("%s: %s: %s\n", args->a0(), fn, strerror(err));
-	 exit_code=1;
-      }
-      break;
-
-   case WRITE:
-      if(int err = lftp_history_write(fn)) {
-	 eprintf("%s: %s: %s\n", args->a0(), fn, strerror(err));
-	 exit_code=1;
-      }
-      break;
-
-   case LIST:
-      lftp_history_list(cnt);
-      break;
-   case CLEAR:
-      lftp_history_clear();
-      break;
-   }
-
-   return 0;
-}
-
 CMD(slot)
 {
    const char *n=args->getarg(1);
@@ -3301,6 +3306,7 @@ CMD(slot)
 CMD(tasks)
 {
    printf("task_count=%d\n",SMTask::TaskCount());
+   SMTask::PrintTasks();
    exit_code=0;
    return 0;
 }
