@@ -1,36 +1,35 @@
 #include <config.h>
-#include "Http.h"
-#include "log.h"
-#include "ResMgr.h"
-#include "SignalHook.h"
 #include <unistd.h>
+#include "FileAccess.h"
+#include "log.h"
 
 int main()
 {
-   ResMgr::ClassInit();
-   SignalHook::ClassInit();
-   Resolver::ClassInit();
-   FileAccess::ClassInit();
-
-   Log::global=new Log();
    Log::global->SetOutput(2,false);
    Log::global->SetLevel(5);
    Log::global->Enable();
+   Log::global->ShowNothing();
 
-   Http *f=new Http;
-   f->Connect("ftp.yars.free.net",0);
-   f->Open("/pub/software/",f->RETRIEVE);
+   FileAccess *f=FileAccess::New("http","ftp.yar.ru");
+   if(!f)
+   {
+      fprintf(stderr,"http: unknown protocol, cannot create http session\n");
+      return 1;
+   }
+   f->Open("/pub/source/lftp/",f->RETRIEVE);
    for(;;)
    {
       SMTask::Schedule();
-      SMTask::Block();
 
       char buf[1024];
       int res=f->Read(buf,sizeof(buf));
       if(res<0)
       {
 	 if(res==f->DO_AGAIN)
+	 {
+	    SMTask::Block();
 	    continue;
+	 }
 	 fprintf(stderr,"Error: %s\n",f->StrError(res));
 	 return 1;
       }
